#include <obs-module.h>
#include <aom/aomcx.h>
#include <util/platform.h>

struct obs_aom {
	obs_encoder_t *encoder;

	aom_codec_enc_cfg_t params;
	aom_codec_ctx_t context;

	aom_image_t img;

	os_performance_token_t *performance_token;

	bool init;
};

static const char *obs_aom_getname(void *unused)
{
	UNUSED_PARAMETER(unused);
	return "AV1";
}

static void clear_data(struct obs_aom *obsaom)
{
	aom_img_free(&obsaom->img);
	aom_codec_destroy(&obsaom->context);
}

static void obs_aom_destroy(void *data)
{
	struct obs_aom *obsaom = data;

	if (obsaom) {
		os_end_high_performance(obsaom->performance_token);
		clear_data(obsaom);
		bfree(obsaom);
	}
}

static void obs_aom_defaults(obs_data_t *settings)
{
	obs_data_set_default_int(settings, "bitrate", 2500);
	obs_data_set_default_int(settings, "buf-sz", 100);
	obs_data_set_default_int(settings, "buf-initial-sz", 100);
	obs_data_set_default_int(settings, "buf-optimal-sz", 100);
	obs_data_set_default_int(settings, "keyint_sec", 5);
	obs_data_set_default_int(settings, "usage", AOM_USAGE_REALTIME);
	obs_data_set_default_int(settings, "cpu-used", 7);
	obs_data_set_default_int(settings, "min-q", 4);
	obs_data_set_default_int(settings, "max-q", 63);
	obs_data_set_default_int(settings, "max-intra-rate", 450);
	obs_data_set_default_int(settings, "max-inter-rate", 0);
	obs_data_set_default_int(settings, "static-thresh", 1);
}

static obs_properties_t *obs_aom_props(void *unused)
{
	UNUSED_PARAMETER(unused);

	obs_properties_t *props = obs_properties_create();
	obs_property_t *p;

	p = obs_properties_add_int(props, "bitrate", obs_module_text("Bitrate"),
				   50, 10000000, 50);
	obs_property_int_set_suffix(p, " Kbps");

	p = obs_properties_add_int(props, "buf-sz", obs_module_text("buf-sz"),
				   1, 200, 1);
	obs_property_int_set_suffix(p, " %");

	p = obs_properties_add_int(props, "buf-initial-sz",
				   obs_module_text("buf-initial-sz"), 1, 200,
				   1);
	obs_property_int_set_suffix(p, " %");

	p = obs_properties_add_int(props, "buf-optimal-sz",
				   obs_module_text("buf-optimal-sz"), 1, 200,
				   1);
	obs_property_int_set_suffix(p, " %");

	obs_properties_add_int(props, "keyint_sec",
			       obs_module_text("KeyframeIntervalSec"), 1, 20,
			       1);

	p = obs_properties_add_list(props, "usage", obs_module_text("Usage"),
				    OBS_COMBO_TYPE_LIST, OBS_COMBO_FORMAT_INT);
	obs_property_list_add_int(p, "realtime", AOM_USAGE_REALTIME);
	obs_property_list_add_int(p, "good", AOM_USAGE_GOOD_QUALITY);

	obs_properties_add_int(props, "cpu-used", obs_module_text("CPUUsed"), 0,
			       11, 1);
	obs_properties_add_int(props, "min-q", obs_module_text("min-q"), 0, 63,
			       1);
	obs_properties_add_int(props, "max-q", obs_module_text("max-q"), 0, 63,
			       1);
	obs_properties_add_int(props, "max-intra-rate",
			       obs_module_text("max-intra-rate"), 0, INT_MAX,
			       1);
	obs_properties_add_int(props, "max-inter-rate",
			       obs_module_text("max-inter-rate"), 0, INT_MAX,
			       1);
	obs_properties_add_int(props, "static-thresh",
			       obs_module_text("static-thresh"), 0, INT_MAX, 1);

	return props;
}

static bool obs_aom_update(void *data, obs_data_t *settings)
{
	aom_codec_err_t res;
	aom_enc_frame_flags_t flags = 0;
	struct obs_aom *obsaom = data;
	video_t *video = obs_encoder_video(obsaom->encoder);
	const struct video_output_info *voi = video_output_get_info(video);
	aom_codec_iface_t *codec_iface = aom_codec_av1_cx();

	unsigned int old_usage;
	if (!obsaom->init)
		old_usage = obsaom->params.g_usage;

	unsigned int usage = (unsigned int)obs_data_get_int(settings, "usage");
	res = aom_codec_enc_config_default(codec_iface, &obsaom->params, usage);
	if (res)
		return false;

	int keyint_sec = (int)obs_data_get_int(settings, "keyint_sec");
	obsaom->params.kf_max_dist = keyint_sec * voi->fps_num / voi->fps_den;
	obsaom->params.rc_target_bitrate =
		(unsigned int)obs_data_get_int(settings, "bitrate");
	obsaom->params.rc_buf_sz =
		obsaom->params.rc_target_bitrate / 100 *
		(unsigned int)obs_data_get_int(settings, "buf-sz");
	obsaom->params.rc_buf_initial_sz =
		obsaom->params.rc_target_bitrate / 100 *
		(unsigned int)obs_data_get_int(settings, "buf-initial-sz");
	obsaom->params.rc_buf_optimal_sz =
		obsaom->params.rc_target_bitrate / 100 *
		(unsigned int)obs_data_get_int(settings, "buf-optimal-sz");
	obsaom->params.g_w = obs_encoder_get_width(obsaom->encoder);
	obsaom->params.g_h = obs_encoder_get_height(obsaom->encoder);
	obsaom->params.g_timebase.num = voi->fps_den;
	obsaom->params.g_timebase.den = voi->fps_num;
	obsaom->params.g_pass = AOM_RC_ONE_PASS;
	obsaom->params.rc_end_usage = AOM_CBR;
	obsaom->params.rc_min_quantizer =
		(unsigned int)obs_data_get_int(settings, "min-q");
	obsaom->params.rc_max_quantizer =
		(unsigned int)obs_data_get_int(settings, "max-q");
	obsaom->params.rc_undershoot_pct = 100;
	obsaom->params.rc_overshoot_pct = 100;
	obsaom->params.g_lag_in_frames = 0;
	obsaom->params.g_error_resilient = 1;

	if (voi->format == VIDEO_FORMAT_I010 ||
	    voi->format == VIDEO_FORMAT_P010) {
		obsaom->params.g_profile = 0;
		obsaom->params.g_bit_depth = AOM_BITS_10;
		obsaom->params.g_input_bit_depth = AOM_BITS_10;
		flags |= AOM_CODEC_USE_HIGHBITDEPTH;
	} else if (voi->format == VIDEO_FORMAT_I444) {
		obsaom->params.g_profile = 1;
		obsaom->params.g_bit_depth = AOM_BITS_8;
		obsaom->params.g_input_bit_depth = AOM_BITS_8;
	} else {
		obsaom->params.g_profile = 0;
		obsaom->params.g_bit_depth = AOM_BITS_8;
		obsaom->params.g_input_bit_depth = AOM_BITS_8;
	}

	int log2_tile_columns = min(
		6,
		max(0, (int)floor(log2(ceil(obsaom->params.g_w / 64.0)) - 2)));
	obsaom->params.g_threads =
		min(os_get_logical_cores(), (int)pow(2, log2_tile_columns + 1));

	if (obsaom->init) {
		res = aom_codec_enc_init(&obsaom->context, codec_iface,
					 &obsaom->params, flags);
	} else if (old_usage == obsaom->params.g_usage) {
		res = aom_codec_enc_config_set(&obsaom->context,
					       &obsaom->params);
	} else {
		aom_codec_destroy(&obsaom->context);
		res = aom_codec_enc_init(&obsaom->context, codec_iface,
					 &obsaom->params, flags);
	}
	if (res)
		return false;

	aom_codec_control(&obsaom->context, AV1E_SET_TILE_COLUMNS, 6);
	aom_codec_control(&obsaom->context, AV1E_SET_TILE_ROWS, 6);
	aom_codec_control(&obsaom->context, AV1E_SET_ROW_MT, 1);

	int max_inter_bitrate_pct =
		(int)obs_data_get_int(settings, "max-inter-rate");
	aom_codec_control(&obsaom->context, AV1E_SET_MAX_INTER_BITRATE_PCT,
			  max_inter_bitrate_pct);

	int cpu_used = (int)obs_data_get_int(settings, "cpu-used");
	aom_codec_control(&obsaom->context, AOME_SET_CPUUSED, cpu_used);

	int static_threshold = (int)obs_data_get_int(settings, "static-thresh");
	aom_codec_control(&obsaom->context, AOME_SET_STATIC_THRESHOLD,
			  static_threshold);

	int max_intra_bitrate_pct =
		(int)obs_data_get_int(settings, "max-intra-rate");
	aom_codec_control(&obsaom->context, AOME_SET_MAX_INTRA_BITRATE_PCT,
			  max_intra_bitrate_pct);

	return false;
}

static void *obs_aom_create(obs_data_t *settings, obs_encoder_t *encoder)
{
	struct obs_aom *obsaom = bzalloc(sizeof(struct obs_aom));
	obsaom->encoder = encoder;

	obsaom->init = true;
	obs_aom_update(obsaom, settings);
	obsaom->init = false;

	video_t *video = obs_encoder_video(obsaom->encoder);
	const struct video_output_info *voi = video_output_get_info(video);
	aom_img_fmt_t img_fmt;

	switch (voi->format) {
	case VIDEO_FORMAT_NV12:
		img_fmt = AOM_IMG_FMT_NV12;
		break;
	case VIDEO_FORMAT_I420:
		img_fmt = AOM_IMG_FMT_I420;
		break;
	case VIDEO_FORMAT_I444:
		img_fmt = AOM_IMG_FMT_I444;
		break;
	case VIDEO_FORMAT_P010:
	case VIDEO_FORMAT_I010:
		img_fmt = AOM_IMG_FMT_I42016;
		break;
	default:
		img_fmt = AOM_IMG_FMT_NV12;
	}

	if (!aom_img_alloc(&obsaom->img, img_fmt, obsaom->params.g_w,
			   obsaom->params.g_h, 1))
		return NULL;

	switch (voi->colorspace) {
	case VIDEO_CS_DEFAULT:
	case VIDEO_CS_709:
		obsaom->img.cp = AOM_CICP_CP_BT_709;
		obsaom->img.tc = AOM_CICP_TC_BT_709;
		obsaom->img.mc = AOM_CICP_MC_BT_709;
		break;
	case VIDEO_CS_601:
		obsaom->img.cp = AOM_CICP_CP_BT_601;
		obsaom->img.tc = AOM_CICP_TC_BT_601;
		obsaom->img.mc = AOM_CICP_MC_BT_601;
		break;
	case VIDEO_CS_SRGB:
		obsaom->img.cp = AOM_CICP_CP_BT_709;
		obsaom->img.tc = AOM_CICP_TC_IEC_61966;
		obsaom->img.mc = AOM_CICP_MC_BT_709;
		break;
	case VIDEO_CS_2100_PQ:
		obsaom->img.cp = AOM_CICP_CP_BT_2020;
		obsaom->img.tc = AOM_CICP_TC_SMPTE_2084;
		obsaom->img.mc = AOM_CICP_MC_BT_2020_NCL;
		break;
	case VIDEO_CS_2100_HLG:
		obsaom->img.cp = AOM_CICP_CP_BT_2020;
		obsaom->img.tc = AOM_CICP_TC_HLG;
		obsaom->img.mc = AOM_CICP_MC_BT_2020_NCL;
	}

	obsaom->img.range = voi->range == VIDEO_RANGE_FULL;

	obsaom->performance_token = os_request_high_performance("aom encoding");

	return obsaom;
}

static bool obs_aom_encode(void *data, struct encoder_frame *frame,
			   struct encoder_packet *packet, bool *received_packet)
{
	struct obs_aom *obsaom = data;

	if (!frame || !packet || !received_packet)
		return false;

	memcpy(obsaom->img.planes[AOM_PLANE_Y], frame->data[0],
	       frame->linesize[0] * (size_t)obsaom->img.h);

	switch (obsaom->img.fmt) {
	case AOM_IMG_FMT_NV12:
		memcpy(obsaom->img.planes[AOM_PLANE_U], frame->data[1],
		       frame->linesize[1] * obsaom->img.h / 2);
		break;
	case AOM_IMG_FMT_I420:
	case AOM_IMG_FMT_I42016:
		memcpy(obsaom->img.planes[AOM_PLANE_U], frame->data[1],
		       frame->linesize[1] * obsaom->img.h / 2);
		memcpy(obsaom->img.planes[AOM_PLANE_V], frame->data[2],
		       frame->linesize[2] * obsaom->img.h / 2);
		break;
	case AOM_IMG_FMT_I444:
		memcpy(obsaom->img.planes[AOM_PLANE_U], frame->data[1],
		       frame->linesize[1] * (size_t)obsaom->img.h);
		memcpy(obsaom->img.planes[AOM_PLANE_V], frame->data[2],
		       frame->linesize[2] * (size_t)obsaom->img.h);
	}

	aom_codec_iter_t iter = NULL;
	const aom_codec_cx_pkt_t *pkt = NULL;
	const aom_codec_err_t res = aom_codec_encode(
		&obsaom->context, &obsaom->img, frame->pts, 1, 0);
	if (res)
		return false;

	while ((pkt = aom_codec_get_cx_data(&obsaom->context, &iter)) != NULL) {
		if (pkt->kind != AOM_CODEC_CX_FRAME_PKT)
			continue;

		packet->data = pkt->data.frame.buf;
		packet->size = pkt->data.frame.sz;
		packet->type = OBS_ENCODER_VIDEO;
		packet->pts = pkt->data.frame.pts;
		packet->dts = pkt->data.frame.pts;
		packet->keyframe = !!(pkt->data.frame.flags & AOM_FRAME_IS_KEY);
		*received_packet = true;
		break;
	}

	return true;
}

static bool obs_aom_extra_data(void *data, uint8_t **extra_data, size_t *size)
{
	struct obs_aom *obsaom = data;
	aom_fixed_buf_t *buf = aom_codec_get_global_headers(&obsaom->context);

	if (!buf)
		return false;

	*extra_data = (uint8_t *)buf->buf;
	*size = buf->sz;
	return true;
}

static void obs_aom_video_info(void *data, struct video_scale_info *info)
{
	struct obs_aom *obsaom = data;
	switch (obsaom->img.fmt) {
	case AOM_IMG_FMT_NV12:
		info->format = VIDEO_FORMAT_NV12;
		break;
	case AOM_IMG_FMT_I420:
		info->format = VIDEO_FORMAT_I420;
		break;
	case AOM_IMG_FMT_I444:
		info->format = VIDEO_FORMAT_I444;
		break;
	case AOM_IMG_FMT_I42016:
		info->format = VIDEO_FORMAT_I010;
		break;
	default:
		info->format = VIDEO_FORMAT_NV12;
	}
}

struct obs_encoder_info obs_aom_encoder = {
	.id = "obs_aom",
	.type = OBS_ENCODER_VIDEO,
	.codec = "av1",
	.get_name = obs_aom_getname,
	.create = obs_aom_create,
	.destroy = obs_aom_destroy,
	.encode = obs_aom_encode,
	.update = obs_aom_update,
	.get_properties = obs_aom_props,
	.get_defaults = obs_aom_defaults,
	.get_extra_data = obs_aom_extra_data,
	.get_video_info = obs_aom_video_info,
	.caps = OBS_ENCODER_CAP_DYN_BITRATE,
};
