#include <obs-module.h>
#include <vpx/vp8cx.h>
#include <util/platform.h>

struct obs_vpx {
	obs_encoder_t *encoder;

	vpx_codec_enc_cfg_t params;
	vpx_codec_ctx_t context;

	vpx_image_t img;

	os_performance_token_t *performance_token;

	unsigned int deadline;
	bool is_vp9;
	bool init;
};

static const char *obs_vp8_getname(void *unused)
{
	UNUSED_PARAMETER(unused);
	return "VP8";
}

static const char *obs_vp9_getname(void *unused)
{
	UNUSED_PARAMETER(unused);
	return "VP9";
}

static void clear_data(struct obs_vpx *obsvpx)
{
	vpx_img_free(&obsvpx->img);
	vpx_codec_destroy(&obsvpx->context);
}

static void obs_vpx_destroy(void *data)
{
	struct obs_vpx *obsvpx = data;

	if (obsvpx) {
		os_end_high_performance(obsvpx->performance_token);
		clear_data(obsvpx);
		bfree(obsvpx);
	}
}

static void obs_vp8_defaults(obs_data_t *settings)
{
	obs_data_set_default_int(settings, "bitrate", 2500);
	obs_data_set_default_int(settings, "buf-sz", 100);
	obs_data_set_default_int(settings, "buf-initial-sz", 100);
	obs_data_set_default_int(settings, "buf-optimal-sz", 100);
	obs_data_set_default_int(settings, "keyint_sec", 5);
	obs_data_set_default_int(settings, "deadline", VPX_DL_REALTIME);
	obs_data_set_default_int(settings, "cpu-used", -3);
	obs_data_set_default_int(settings, "min-q", 4);
	obs_data_set_default_int(settings, "max-q", 63);
	obs_data_set_default_int(settings, "max-intra-rate", 450);
	obs_data_set_default_int(settings, "static-thresh", 1);
}

static void obs_vp9_defaults(obs_data_t *settings)
{
	obs_data_set_default_int(settings, "bitrate", 2500);
	obs_data_set_default_int(settings, "buf-sz", 100);
	obs_data_set_default_int(settings, "buf-initial-sz", 100);
	obs_data_set_default_int(settings, "buf-optimal-sz", 100);
	obs_data_set_default_int(settings, "keyint_sec", 5);
	obs_data_set_default_int(settings, "deadline", VPX_DL_REALTIME);
	obs_data_set_default_int(settings, "cpu-used", 5);
	obs_data_set_default_int(settings, "min-q", 4);
	obs_data_set_default_int(settings, "max-q", 63);
	obs_data_set_default_int(settings, "max-intra-rate", 450);
	obs_data_set_default_int(settings, "max-inter-rate", 0);
	obs_data_set_default_int(settings, "static-thresh", 1);
}

static obs_properties_t *obs_vp8_props(void *unused)
{
	UNUSED_PARAMETER(unused);

	obs_properties_t *props = obs_properties_create();
	obs_property_t *p;

	p = obs_properties_add_int(props, "bitrate", obs_module_text("Bitrate"),
				   50, 10000000, 50);
	obs_property_int_set_suffix(p, " Kbps");

	p = obs_properties_add_int(props, "buf-sz", obs_module_text("buf-sz"),
				   1, 200, 1);
	obs_property_int_set_suffix(p, " %");

	p = obs_properties_add_int(props, "buf-initial-sz",
				   obs_module_text("buf-initial-sz"), 1, 200,
				   1);
	obs_property_int_set_suffix(p, " %");

	p = obs_properties_add_int(props, "buf-optimal-sz",
				   obs_module_text("buf-optimal-sz"), 1, 200,
				   1);
	obs_property_int_set_suffix(p, " %");

	obs_properties_add_int(props, "keyint_sec",
			       obs_module_text("KeyframeIntervalSec"), 1, 20,
			       1);

	p = obs_properties_add_list(props, "deadline",
				    obs_module_text("DeadLine"),
				    OBS_COMBO_TYPE_LIST, OBS_COMBO_FORMAT_INT);
	obs_property_list_add_int(p, "realtime", VPX_DL_REALTIME);
	obs_property_list_add_int(p, "good", VPX_DL_GOOD_QUALITY);
	obs_property_list_add_int(p, "best", VPX_DL_BEST_QUALITY);

	obs_properties_add_int(props, "cpu-used", obs_module_text("CPUUsed"),
			       -16, 16, 1);
	obs_properties_add_int(props, "min-q", obs_module_text("min-q"), 0, 63,
			       1);
	obs_properties_add_int(props, "max-q", obs_module_text("max-q"), 0, 63,
			       1);
	obs_properties_add_int(props, "max-intra-rate",
			       obs_module_text("max-intra-rate"), 0, INT_MAX,
			       1);
	obs_properties_add_int(props, "static-thresh",
			       obs_module_text("static-thresh"), 0, INT_MAX, 1);

	return props;
}

static obs_properties_t *obs_vp9_props(void *unused)
{
	UNUSED_PARAMETER(unused);

	obs_properties_t *props = obs_properties_create();
	obs_property_t *p;

	p = obs_properties_add_int(props, "bitrate", obs_module_text("Bitrate"),
				   50, 10000000, 50);
	obs_property_int_set_suffix(p, " Kbps");

	p = obs_properties_add_int(props, "buf-sz", obs_module_text("buf-sz"),
				   1, 200, 1);
	obs_property_int_set_suffix(p, " %");

	p = obs_properties_add_int(props, "buf-initial-sz",
				   obs_module_text("buf-initial-sz"), 1, 200,
				   1);
	obs_property_int_set_suffix(p, " %");

	p = obs_properties_add_int(props, "buf-optimal-sz",
				   obs_module_text("buf-optimal-sz"), 1, 200,
				   1);
	obs_property_int_set_suffix(p, " %");

	obs_properties_add_int(props, "keyint_sec",
			       obs_module_text("KeyframeIntervalSec"), 1, 20,
			       1);

	p = obs_properties_add_list(props, "deadline",
				    obs_module_text("DeadLine"),
				    OBS_COMBO_TYPE_LIST, OBS_COMBO_FORMAT_INT);
	obs_property_list_add_int(p, "realtime", VPX_DL_REALTIME);
	obs_property_list_add_int(p, "good", VPX_DL_GOOD_QUALITY);
	obs_property_list_add_int(p, "best", VPX_DL_BEST_QUALITY);

	obs_properties_add_int(props, "cpu-used", obs_module_text("CPUUsed"),
			       -9, 9, 1);
	obs_properties_add_int(props, "min-q", obs_module_text("min-q"), 0, 63,
			       1);
	obs_properties_add_int(props, "max-q", obs_module_text("max-q"), 0, 63,
			       1);
	obs_properties_add_int(props, "max-intra-rate",
			       obs_module_text("max-intra-rate"), 0, INT_MAX,
			       1);
	obs_properties_add_int(props, "max-inter-rate",
			       obs_module_text("max-inter-rate"), 0, INT_MAX,
			       1);
	obs_properties_add_int(props, "static-thresh",
			       obs_module_text("static-thresh"), 0, INT_MAX, 1);

	return props;
}

static bool obs_vpx_update(void *data, obs_data_t *settings)
{
	vpx_codec_err_t res;
	vpx_enc_frame_flags_t flags = 0;
	struct obs_vpx *obsvpx = data;
	video_t *video = obs_encoder_video(obsvpx->encoder);
	const struct video_output_info *voi = video_output_get_info(video);
	vpx_codec_iface_t *codec_iface = obsvpx->is_vp9 ? vpx_codec_vp9_cx()
							: vpx_codec_vp8_cx();

	unsigned int old_deadline;
	if (!obsvpx->init)
		old_deadline = obsvpx->deadline;

	obsvpx->deadline = (unsigned int)obs_data_get_int(settings, "deadline");
	res = vpx_codec_enc_config_default(codec_iface, &obsvpx->params, 0);
	if (res)
		return false;

	int keyint_sec = (int)obs_data_get_int(settings, "keyint_sec");
	obsvpx->params.kf_max_dist = keyint_sec * voi->fps_num / voi->fps_den;
	obsvpx->params.rc_target_bitrate =
		(unsigned int)obs_data_get_int(settings, "bitrate");
	obsvpx->params.rc_buf_sz =
		obsvpx->params.rc_target_bitrate / 100 *
		(unsigned int)obs_data_get_int(settings, "buf-sz");
	obsvpx->params.rc_buf_initial_sz =
		obsvpx->params.rc_target_bitrate / 100 *
		(unsigned int)obs_data_get_int(settings, "buf-initial-sz");
	obsvpx->params.rc_buf_optimal_sz =
		obsvpx->params.rc_target_bitrate / 100 *
		(unsigned int)obs_data_get_int(settings, "buf-optimal-sz");
	obsvpx->params.g_w = obs_encoder_get_width(obsvpx->encoder);
	obsvpx->params.g_h = obs_encoder_get_height(obsvpx->encoder);
	obsvpx->params.g_timebase.num = voi->fps_den;
	obsvpx->params.g_timebase.den = voi->fps_num;
	obsvpx->params.g_pass = VPX_RC_ONE_PASS;
	obsvpx->params.rc_end_usage = VPX_CBR;
	obsvpx->params.rc_min_quantizer =
		(unsigned int)obs_data_get_int(settings, "min-q");
	obsvpx->params.rc_max_quantizer =
		(unsigned int)obs_data_get_int(settings, "max-q");
	obsvpx->params.rc_undershoot_pct = 100;
	obsvpx->params.rc_overshoot_pct = 100;
	obsvpx->params.g_lag_in_frames = 0;
	obsvpx->params.g_error_resilient = 1;

	if (obsvpx->is_vp9 && (voi->format == VIDEO_FORMAT_I010 ||
			       voi->format == VIDEO_FORMAT_P010)) {
		obsvpx->params.g_profile = 2;
		obsvpx->params.g_bit_depth = VPX_BITS_10;
		obsvpx->params.g_input_bit_depth = VPX_BITS_10;
		flags |= VPX_CODEC_USE_HIGHBITDEPTH;
	} else if (obsvpx->is_vp9 && voi->format == VIDEO_FORMAT_I444) {
		obsvpx->params.g_profile = 1;
		obsvpx->params.g_bit_depth = VPX_BITS_8;
		obsvpx->params.g_input_bit_depth = VPX_BITS_8;
	} else {
		obsvpx->params.g_profile = 0;
		obsvpx->params.g_bit_depth = VPX_BITS_8;
		obsvpx->params.g_input_bit_depth = VPX_BITS_8;
	}

	int log2_tile_columns = min(
		6,
		max(0, (int)floor(log2(ceil(obsvpx->params.g_w / 64.0)) - 2)));
	obsvpx->params.g_threads =
		min(os_get_logical_cores(), (int)pow(2, log2_tile_columns + 1));

	if (obsvpx->init) {
		res = vpx_codec_enc_init(&obsvpx->context, codec_iface,
					 &obsvpx->params, flags);
	} else if (old_deadline == obsvpx->deadline) {
		res = vpx_codec_enc_config_set(&obsvpx->context,
					       &obsvpx->params);
	} else {
		vpx_codec_destroy(&obsvpx->context);
		res = vpx_codec_enc_init(&obsvpx->context, codec_iface,
					 &obsvpx->params, flags);
	}
	if (res)
		return false;

	if (obsvpx->is_vp9) {
		vpx_codec_control(&obsvpx->context, VP9E_SET_TILE_COLUMNS, 6);
		vpx_codec_control(&obsvpx->context, VP9E_SET_TILE_ROWS, 6);
		vpx_codec_control(&obsvpx->context, VP9E_SET_ROW_MT, 1);
		vpx_codec_control(&obsvpx->context,
				  VP9E_SET_DISABLE_OVERSHOOT_MAXQ_CBR, 1);

		int max_inter_bitrate_pct =
			(int)obs_data_get_int(settings, "max-inter-rate");
		vpx_codec_control(&obsvpx->context,
				  VP9E_SET_MAX_INTER_BITRATE_PCT,
				  max_inter_bitrate_pct);
	}

	int cpu_used = (int)obs_data_get_int(settings, "cpu-used");
	vpx_codec_control(&obsvpx->context, VP8E_SET_CPUUSED, cpu_used);

	int static_threshold = (int)obs_data_get_int(settings, "static-thresh");
	vpx_codec_control(&obsvpx->context, VP8E_SET_STATIC_THRESHOLD,
			  static_threshold);

	int max_intra_bitrate_pct =
		(int)obs_data_get_int(settings, "max-intra-rate");
	vpx_codec_control(&obsvpx->context, VP8E_SET_MAX_INTRA_BITRATE_PCT,
			  max_intra_bitrate_pct);

	return false;
}

static bool obs_vpx_create(struct obs_vpx *obsvpx, obs_data_t *settings)
{
	obsvpx->init = true;
	obs_vpx_update(obsvpx, settings);
	obsvpx->init = false;

	video_t *video = obs_encoder_video(obsvpx->encoder);
	const struct video_output_info *voi = video_output_get_info(video);
	vpx_img_fmt_t img_fmt;

	switch (voi->format) {
	case VIDEO_FORMAT_NV12:
		img_fmt = VPX_IMG_FMT_NV12;
		break;
	case VIDEO_FORMAT_I420:
		img_fmt = VPX_IMG_FMT_I420;
		break;
	case VIDEO_FORMAT_I444:
		if (obsvpx->is_vp9) {
			img_fmt = VPX_IMG_FMT_I444;
			break;
		}
	case VIDEO_FORMAT_P010:
	case VIDEO_FORMAT_I010:
		if (obsvpx->is_vp9) {
			img_fmt = VPX_IMG_FMT_I42016;
			break;
		}
	default:
		img_fmt = VPX_IMG_FMT_NV12;
	}

	if (!vpx_img_alloc(&obsvpx->img, img_fmt, obsvpx->params.g_w,
			   obsvpx->params.g_h, 1))
		return false;

	switch (voi->colorspace) {
	case VIDEO_CS_DEFAULT:
	case VIDEO_CS_709:
		obsvpx->img.cs = VPX_CS_BT_709;
		break;
	case VIDEO_CS_601:
		obsvpx->img.cs = VPX_CS_BT_601;
		break;
	case VIDEO_CS_SRGB:
		obsvpx->img.cs = VPX_CS_SRGB;
		break;
	case VIDEO_CS_2100_PQ:
	case VIDEO_CS_2100_HLG:
		obsvpx->img.cs = VPX_CS_BT_2020;
	}

	obsvpx->img.range = voi->range == VIDEO_RANGE_FULL;

	obsvpx->performance_token = os_request_high_performance("vpx encoding");

	return true;
}

static void *obs_vp8_create(obs_data_t *settings, obs_encoder_t *encoder)
{
	struct obs_vpx *obsvpx = bzalloc(sizeof(struct obs_vpx));
	obsvpx->encoder = encoder;
	obsvpx->is_vp9 = false;

	return obs_vpx_create(obsvpx, settings) ? obsvpx : NULL;
}

static void *obs_vp9_create(obs_data_t *settings, obs_encoder_t *encoder)
{
	struct obs_vpx *obsvpx = bzalloc(sizeof(struct obs_vpx));
	obsvpx->encoder = encoder;
	obsvpx->is_vp9 = true;

	return obs_vpx_create(obsvpx, settings) ? obsvpx : NULL;
}

static bool obs_vpx_encode(void *data, struct encoder_frame *frame,
			   struct encoder_packet *packet, bool *received_packet)
{
	struct obs_vpx *obsvpx = data;

	if (!frame || !packet || !received_packet)
		return false;

	memcpy(obsvpx->img.planes[VPX_PLANE_Y], frame->data[0],
	       frame->linesize[0] * (size_t)obsvpx->img.h);

	switch (obsvpx->img.fmt) {
	case VPX_IMG_FMT_NV12:
		memcpy(obsvpx->img.planes[VPX_PLANE_U], frame->data[1],
		       frame->linesize[1] * obsvpx->img.h / 2);
		break;
	case VPX_IMG_FMT_I420:
	case VPX_IMG_FMT_I42016:
		memcpy(obsvpx->img.planes[VPX_PLANE_U], frame->data[1],
		       frame->linesize[1] * obsvpx->img.h / 2);
		memcpy(obsvpx->img.planes[VPX_PLANE_V], frame->data[2],
		       frame->linesize[2] * obsvpx->img.h / 2);
		break;
	case VPX_IMG_FMT_I444:
		memcpy(obsvpx->img.planes[VPX_PLANE_U], frame->data[1],
		       frame->linesize[1] * (size_t)obsvpx->img.h);
		memcpy(obsvpx->img.planes[VPX_PLANE_V], frame->data[2],
		       frame->linesize[2] * (size_t)obsvpx->img.h);
	}

	vpx_codec_iter_t iter = NULL;
	const vpx_codec_cx_pkt_t *pkt = NULL;
	const vpx_codec_err_t res = vpx_codec_encode(&obsvpx->context,
						     &obsvpx->img, frame->pts,
						     1, 0, obsvpx->deadline);
	if (res)
		return false;

	while ((pkt = vpx_codec_get_cx_data(&obsvpx->context, &iter)) != NULL) {
		if (pkt->kind != VPX_CODEC_CX_FRAME_PKT)
			continue;

		packet->data = pkt->data.frame.buf;
		packet->size = pkt->data.frame.sz;
		packet->type = OBS_ENCODER_VIDEO;
		packet->pts = pkt->data.frame.pts;
		packet->dts = pkt->data.frame.pts;
		packet->keyframe = !!(pkt->data.frame.flags & VPX_FRAME_IS_KEY);
		*received_packet = true;
		break;
	}

	return true;
}

static bool obs_vpx_extra_data(void *data, uint8_t **extra_data, size_t *size)
{
	*extra_data = NULL;
	*size = 0;
	return true;
}

static void obs_vpx_video_info(void *data, struct video_scale_info *info)
{
	struct obs_vpx *obsvpx = data;
	switch (obsvpx->img.fmt) {
	case VPX_IMG_FMT_NV12:
		info->format = VIDEO_FORMAT_NV12;
		break;
	case VPX_IMG_FMT_I420:
		info->format = VIDEO_FORMAT_I420;
		break;
	case VPX_IMG_FMT_I444:
		info->format = VIDEO_FORMAT_I444;
		break;
	case VPX_IMG_FMT_I42016:
		info->format = VIDEO_FORMAT_I010;
		break;
	default:
		info->format = VIDEO_FORMAT_NV12;
	}
}

struct obs_encoder_info obs_vp8_encoder = {
	.id = "obs_vp8",
	.type = OBS_ENCODER_VIDEO,
	.codec = "vp8",
	.get_name = obs_vp8_getname,
	.create = obs_vp8_create,
	.destroy = obs_vpx_destroy,
	.encode = obs_vpx_encode,
	.update = obs_vpx_update,
	.get_properties = obs_vp8_props,
	.get_defaults = obs_vp8_defaults,
	.get_extra_data = obs_vpx_extra_data,
	.get_video_info = obs_vpx_video_info,
	.caps = OBS_ENCODER_CAP_DYN_BITRATE,
};

struct obs_encoder_info obs_vp9_encoder = {
	.id = "obs_vp9",
	.type = OBS_ENCODER_VIDEO,
	.codec = "vp9",
	.get_name = obs_vp9_getname,
	.create = obs_vp9_create,
	.destroy = obs_vpx_destroy,
	.encode = obs_vpx_encode,
	.update = obs_vpx_update,
	.get_properties = obs_vp9_props,
	.get_defaults = obs_vp9_defaults,
	.get_extra_data = obs_vpx_extra_data,
	.get_video_info = obs_vpx_video_info,
	.caps = OBS_ENCODER_CAP_DYN_BITRATE,
};
